/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package cebit.server.idm;

import idman.gui.LocalManagementGUI;
import idman.mngt.impl.LocalManagement;

import java.net.URL;
import java.util.Properties;

import javax.swing.JFrame;

import org.ssonet.util.Preload;
import org.w3c.jigsaw.daemon.ServerHandlerInitException;
import org.w3c.jigsaw.http.httpd;
import org.w3c.tools.resources.*;

import cebit.server.Server;
import cebit.server.ServerProperties;

/**
 * This class is the daemon for handling and serving IDM connections. This daemon
 * also initializes the IDM engine.
 */
public class IDMd extends httpd{
    public static boolean debug = true;

    /** version of this IDMd */
    private static String version = "0.2";

    static {
        // preload security provider and mechanisms for SSONET
        Preload preload = new Preload();
        try{
            preload.join();
        }catch (Exception e){};
        try{
            idman.util.KeyProvider.start();
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }

    /**
     * constructor of this daemon
     */
    public IDMd() {
        super();
    }

    /**
     * clone method of this daemon
     */
    protected Object clone()
        throws CloneNotSupportedException {
        IDMd daemon = (IDMd)(super.clone());
        // additional IDMd's use the same idmanager
        return daemon;
    }

    /**
     * method for supplying the banner of this daemon
     * @return banner of this daemon
     */
    protected String getBanner() {
        return super.getBanner() + "/IDM[" + version +"]";
    }

    /**
     * method for registering IDM property management
     */
    protected void initializePropertySets() {
        super.initializePropertySets();
        registerPropertySet(new ServerProperties(this));
    }

    /**
     * method for initializing the properties of this daemon
     * @exception ServerHandlerInitException if initialization fails
     */
    protected void initializeProperties() throws ServerHandlerInitException {
        super.initializeProperties();
        // initialize IDM
        if(Server.idManager == null){
            Properties settings = new Properties();
            settings.setProperty("dbi.username",
                props.getString(ServerProperties.DBI_USERNAME_P, ""));
            settings.setProperty("dbi.hostname",
                props.getString(ServerProperties.DBI_HOSTNAME_P, ""));
            settings.setProperty("dbi.filename",
                props.getString(ServerProperties.DBI_FILENAME_P, ""));
            settings.setProperty("dbi.password",
                props.getString(ServerProperties.DBI_PASSWORD_P, ""));
            settings.setProperty("dbi.database",
                props.getString(ServerProperties.DBI_DATABASE_P, ""));
            settings.setProperty("dbi.port",
                props.getString(ServerProperties.DBI_PORT_P, ""));
            settings.setProperty("dbi.driverClass",
                props.getString(ServerProperties.DBI_DRIVERCLASS_P, ""));
            settings.setProperty("dbi.conceal",
                props.getString(ServerProperties.DBI_CONCEAL_P, ""));
            settings.setProperty("dbi.passphrase",
                props.getString(ServerProperties.DBI_PASSPHRASE_P, ""));
            Server.idManager  = new LocalManagement(settings);
            if (false) try {
                // create IDManager GUI
                JFrame frame = new JFrame("JigSaw IDManagement");
                Server.idManagerGUI = new LocalManagementGUI(Server.idManager, frame, settings);
                frame.getContentPane().add(Server.idManagerGUI);
                frame.setSize(400, 400);
                frame.setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
                frame.show();
                if(debug) System.out.println("INF: IDMd: IDManager succesfully started");
            } catch (Error e) {
                System.err.println("ERR: Could not open ID-Management GUI: "+e.getMessage());
            }
            // set personal pseudonym
            Server.idManager.settings.setProperty("mngt.certificate.subjectCN",
                props.getString(ServerProperties.IDM_PERSONAL_PSEUDONYM_P, ""));
        }
    }

    /**
     * method for handling a dynamic property modification
     * @param name the name of the property modified
     * @return true if and only if the modification has been handled successfully
     */
    public boolean propertyChanged(String name) {
        if (super.propertyChanged(name)) {
            // handle dynamically changable IDMd properties
            if (name.equals(ServerProperties.IDM_PERSONAL_PSEUDONYM_P)) {
                Server.idManager.settings.setProperty("mngt.certificate.subjectCN",
                props.getString(ServerProperties.IDM_PERSONAL_PSEUDONYM_P,""));
            }
            // TODO: Make more Properties changeable
           return true;
        }
        else return false;
    }

    /**
     * method for supplying a reply interface of a request
     * @param req current request to be handled
     * @return reply for a current request
     */
    public ReplyInterface perform(RequestInterface req) throws ProtocolException, ResourceException  {
        // perform the request
        return super.perform(req);
    }

    /**
     * method for supplying the uri of this daemon
     * @return uri of this daemon
     */
    public URL getURL() {
        return super.getURL();
    }
}
