/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package cebit.server.idm.socket;

import idman.mngt.Context;

import java.io.BufferedInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;

import org.ssonet.net.*;
import org.ssonet.net.ConnectionListener;
import org.ssonet.net.SSONETContext;
import org.ssonet.net.SSONETContextListener;
import org.ssonet.net.impl.SSONETServerClientSocket;
import org.w3c.jigsaw.http.*;
import org.w3c.tools.resources.ServerInterface;

import cebit.server.Server;

/**
 * Implements a Client (Connection handler) for Connections using the
 * Identity manager. Its funktionality is almost the same as the
 * http-SocketClient. The main addition to the basic SocketClient is, that
 * this IDMSocketClient requests a new negotiation of security parameters
 * before the reply is sent. Therefore the SSONETContext of this connection
 * is added to the reply, when calling the method perform(request).
 */
public class IDMSocketClient extends Client implements Runnable, SSONETContextListener,
    ConnectionListener
    {
    private static final boolean trace = false;
    
    /**The ClientFactory that created this client.*/
    private IDMSocketClientFactory pool = null ;
    /**The socket currently handled by the client.*/
    protected Socket socket = null ;
    /** Is this client still alive ?*/
    protected boolean alive = false;
    /**The client state for this client, has managed by the SocketClientFactory*/
    IDMSocketClientState state = null;
    /**Number of times this client was bound to a connection.*/
    protected int bindcount = 0;
    /**The thread that we have been attached to.*/
    protected Thread thread = null;
    /** Our reusable output buffer.*/
    protected SocketOutputBuffer bufout = null;
    /**Are we idle (waiting for next request ?)*/
    protected boolean idle = false;
    /**are we done?*/
    protected boolean done = false;

    protected Context context = null;
    protected boolean isContextRegistered = false;
    protected DataOutputStream dataOutputStream = null;

    /**
     * Create an empty client, that will be ready to work.
     * @param server The server to which this client is attached.
     * @param pool The pool for reuse.
     * @param state The state of this client.
     */
    public IDMSocketClient(httpd server,
                 IDMSocketClientFactory pool,
                 IDMSocketClientState state){
        debug = false;

        initialize(server, state.id);
        this.socket   = null ;
        this.pool     = pool;
        this.state    = state;
        this.alive    = true;
    }

//------------------------------------------------------------------------------
// Methods inherited from abstract class Client

    /**
     * Emit the given reply to the client.
     * @param reply The reply to be emited.
     * @return The number of body bytes emited, or <strong>-1</strong> if
     * no bytes needed to be emitted.
     * @exception IOException If some IO error occurs.
     */

    protected int emitReply (Reply reply)
        throws IOException
    {
        if(debug) System.out.println("INF: IDMSocketClient: start new negotiation...");
        // negotiate the security settings for sending the reply
        //if(context != null) context.doSecurityGoalNegotiation(); XXXX spter wieder rein.....
        // (Exceptions should be correct, if the negotiation results in a conflict)

        // emit the reply
        if(debug) System.out.println("INF: IDMSocketClient: try writing Reply...");
        int written = super.emitReply(reply);
        if(debug) System.out.println("INF: IDMSocketClient: Reply ("+written+" Bytes) emitted.");
        dataOutputStream.flush();
        return written ;
    }


//------------------------------------------------------------------------------
// Methods adapted from SocketClient

    /**
     * Print that client into a String.
     * This adds "IDM" to the output of the toString()-Method of
     * <code>SocketClient</code>.
     * @return A String instance.
     */

    public String toString() {
        if ( thread != null )
            return "IDM-client-"+state.id+"("+thread.getName()+")";
        else
            return "IDM-client-"+state.id;
    }

    /**
     * If this client is allocated a thread, join it.
     */
    public void join() {
        if (thread != null) {
            while ( true ) {
                try {
                    thread.join();
                } catch (InterruptedException ex) {
                }
            }
        }
    }

    /**
     * Run for our newly attached connection.
     * @return A boolean, <strong>true</strong> if the client is to be killed
     * as decided by its SocketClientFactory creator.
     */
    public void run() {
        thread = Thread.currentThread();
        if ( trace )
            System.out.println(this+": powered by "+thread);
        try {
            if ( bufout == null ) {
                // Make sure that buffer is a little smaller then the client
                // buffer. so writing to it will not copy it
                int bufsize = getServer().getClientBufferSize() - 1;
                bufout = new SocketOutputBuffer(socket.getOutputStream()
                                                , bufsize);
            } else {
                bufout.reuse(socket.getOutputStream());
            }
            dataOutputStream = new DataOutputStream(bufout);
            startConnection(new BufferedInputStream(socket.getInputStream())
                            , dataOutputStream);
        } catch (IOException ex) {
            if ( debug )
                ex.printStackTrace();
        } catch (ClientException ex) {
            if ( debug )
                ex.printStackTrace();
            // Emit some debugging traces:
            if ( debug ) {
                if (ex.ex != null )
                    ex.ex.printStackTrace() ;
                else
                    ex.printStackTrace();
            }
            // If output is null, we have killed the connection...
            if ( alive && ! idle ) {
                error("caught ClientException: ["
                      + ex.getClass().getName()
                      + "] " + ex.getMessage()) ;
            }
        } catch (Exception ex) {
            if (debug) {
                System.out.println("ERR: unknown exception caught in client run");
                ex.printStackTrace();
            }
        } finally {
            // update connection pool
            if ( ! pool.clientConnectionFinished(this) ) {
                pool.clientFinished(this);
            }
            thread = null;
        }
    }

    /**
     * Client implementation - Get the IP address of this client.
     * @return An InetAddress instance, or <strong>null</strong> if the
     * client is not currently running.
     */

    public InetAddress getInetAddress () {
        return (socket != null) ? socket.getInetAddress() : null;
    }

    /**
     * Client implementation - This connection has been stopped.
     * Make sure the whole socket is closed, and be ready to handle
     * next connection.
     */
    protected void stopConnection() {
        if ( trace )
            System.out.println(this+": stopConnection.");
        if ( socket != null ) {
            try {
                dataOutputStream.flush();
                if(trace)System.out.println(this+": Outputstream flushed.");
                dataOutputStream.close();
                socket.close();
            } catch (Exception ex) {
            }
            socket = null;
        }else{
            if(trace)System.out.println(this+": socket is null. Not flushed.");
        }
    }

    /**
     * Get the thread powering that client.
     * @return A Thread instance, or <strong>null</strong>.
     */
    protected Thread getThread() {
        return thread;
    }

    /**
     * Client implementation - The current connection is idle.
     * The client is about to wait for the next request from the net, mark
     * it as idle to make it a candidate for persistent connection closing.
     * @return A boolean, if <strong>true</strong> our client factory wants
     * us to stop right now, in order to handle some other connection.
     */
    protected boolean idleConnection() {
        synchronized (state) {
            if ( trace )
                System.out.println(this+": idleConnection.");
            idle = true;
            return ! pool.notifyIdle(this);
        }
    }

    /**
     * Client implementation - The current connection is in use.
     * A request is about to be processed, mark that connection as used, to
     * remove it from the idle state.
     */
    protected void usedConnection() {
        synchronized (state) {
            if ( trace )
                System.out.println(this+": usedConnection.");
            idle = false;
            pool.notifyUse(this) ;
        }
    }

    /**
     * SocketClientFactory interface - Bind the socket to this client.
     * Binding a socket to a client triggers the processing of the underlying
     * connection. It is assumed that the client was ready to handle a new
     * connection before this method was called.
     * @param socket The socket this client should now handle.
     */
    protected synchronized void bind(Socket socket) {
        done = false;
        ServerInterface server = getServer();
        if ( trace )
            System.out.println(this+": bind.");
        this.socket = socket ;
        // register SSONETContextListener
        this.context = (Context)((SSONETServerClientSocket)socket).getContext();
        context.addSSONETContextListener(this);
        context.addConnectionListener(this);
        context.setTargetAddress(socket.getInetAddress());
        context.setTargetPort(socket.getPort());
        // set SoTimeout
        try {
            socket.setSoTimeout(server.getRequestTimeOut());
        } catch (SocketException ex) {
            if (trace)
                ex.printStackTrace();
            server.errlog("IDMSocketClient.bind: Unable to set socket timeout!");
        }
        this.idle   = false;
        bindcount++;
        pool.run(this);
    }

    /**
     * SocketClientFactory interface - Unbind this client.
     * This client is handling an idle connection, unbind it to make
     * it free for handling some other more buzy connection.
     */
    protected synchronized void unbind() {
        if ( trace )
            System.out.println(this+": unbind.");
        interruptConnection(true);
        if (context!=null) {
            context.removeSSONETContextListener(this);
            context.removeConnectionListener(this);
        }
    }

    /**
     * SocketClientFactory interface - Kill this client.
     * The clean way for our client factory to shut us down unconditionally.
     * This will free all resources acquired by this client, stop the current
     * connection processing if needed, and terminate the underlying thread.
     */
    protected synchronized void kill(boolean now) {
        if ( trace )
            System.out.println(this+": kill.");
        alive = false;
        interruptConnection(now);
    }
    /**
     * Get the total number of times this client was bound to a socket.
     * @return An integer, indicatingthe number of bind calls on this client.
     */
    public final int getBindCount() {
        return bindcount;
    }

    protected Request getNextRequest () 
	throws ClientException, IOException
    {
        Request result = super.getNextRequest();
        if (result!=null) {
                result.setState("org.ssonet.context", getContext());
        }
        return result;
    }
                
//----------------------------------------------------------------------
// Methods for the implemented listeners

    /**
     * Method of SSONETContextListener. Replaces an old context.
     */
    public void replaceContext(SSONETContext o, SSONETContext n){
        context = (Context)n;
        o.removeConnectionListener(this);
        n.addConnectionListener(this);
    }

    /**
     * Notifies about context changes. Not used.
     */
    public void contextChanged(SSONETContext c){
        return;
    }

    /**
     * When the connection state has changed, this method reports the current state.
     */
    public void connectionStateChanged(SSONETContext context, int currentState){
        //debug = true;
        switch (currentState){
            case ConnectionListener.RUNNING_STATE:
                // register the context with the IDManager
                if(!isContextRegistered){
                    Server.idManager.registerContext(this.context);
                    isContextRegistered = true;
                    if (debug) System.out.println("IDMSocketClient: Context registered for: "+this);
                }
                break;
            case ConnectionListener.FINISHED_STATE:
                // unregister the context with the IDManager
                if(isContextRegistered){
                    Server.idManager.unregisterContext(this.context);
                    isContextRegistered = false;
                    if (debug) System.out.println("IDMSocketClient: Context unregistered for: "+this);
                }
                break;
        }
    }

    /**
     * @return The socket of this client.
     */
    public Socket getSocket(){
        return socket;
    }

    public Context getContext() {
        return context;
    }

	/* (non-Javadoc)
	 * @see org.ssonet.net.ConnectionListener#negotiationRequestReceived(org.ssonet.net.SSONETContext, int)
	 */
	public void negotiationRequestReceived(SSONETContext context, int negotiationType) {
	}

	/* (non-Javadoc)
	 * @see org.ssonet.net.ConnectionListener#decideSecurityGoal(int, org.ssonet.net.ActionConfiguration, org.ssonet.net.SSONETContext)
	 */
	public boolean decideSecurityGoal(int securityGoal, ActionConfiguration ownConfiguration, SSONETContext context) {
		return false;
	}

}
