/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package cebit.server.idm.socket;

import idman.mngt.Context;
import idman.rules.*;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;

import org.ssonet.net.*;
import org.ssonet.net.impl.*;
import org.w3c.jigsaw.config.PropertySet;
import org.w3c.jigsaw.http.ClientFactory;
import org.w3c.jigsaw.http.httpd;
import org.w3c.util.*;

import cebit.server.Server;

class DebugThread extends Thread {
    IDMSocketClientFactory pool = null;

    public void run() {
        while ( true ) {
            try {
                sleep(1000*10);
                // Display some client statistics:
                IDMSocketClientState cs = null;
                cs = (IDMSocketClientState) pool.freeList.getHead();
                while (cs != null) {
                    System.out.println(cs.client
                                       + " reqcount="
                                       + cs.client.getRequestCount()
                                       + ", bindcount="
                                       + cs.client.getBindCount());
                    cs = (IDMSocketClientState)pool.freeList.getNext((LRUAble)cs);
                }
                System.out.println("freeCount ="+pool.freeCount);
                System.out.println("idleCount ="+pool.idleCount);
                System.out.println("totalCount="+pool.clientCount);
                System.out.println("estimCount="+pool.clientEstim);
                System.out.println("Average: "+pool.loadavg);
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
    }

    DebugThread(IDMSocketClientFactory pool) {
        this.pool = pool;
        setPriority(Thread.MAX_PRIORITY);
    }
}

/**
 * This class handles the SSONET connections to this server, and it creates
 * a ServerSocket for receiving connections.
 * <p>The client pool is a kind of client factory.
 * Each time the server gets a new connection, it calls the client pool
 * to bound a client object (newly created or spared) to handle it.
 */
public class IDMSocketClientFactory implements ClientFactory,
    PropertyMonitoring, Status, SSONETContextFactory, ChainEvaluationListener,
    ConnectionListener
     {

    private static final boolean debug       = false;

    /** static basic factory for the SocketClientFactory for SSONET connections */
    private static SSONETServerSocketFactory factory = null;
    /** The default certificate Listener used for all connections */
    protected CertificateListener certificateListener;
    /** the daemon (Server) of this factory */
    httpd server = null;
    /** the daemon (Server) bind address for this factory */
    InetAddress bindAddr = null;
    /** the maximum number of clients */
    int maxClients = 0;

    /**
     * default constructor of this factory
     */
    public IDMSocketClientFactory() {
    }

    /**
     * method for creating a SSONET server socket
     * @return a new server socket instance
     */
    public ServerSocket createServerSocket() throws IOException {
        int port = server.getPort();
        int clients = Math.max(128, maxClients);
        ServerSocket serversocket = null;
        if (bindAddr == null){
            serversocket = factory.createServerSocket(port, clients);
        }else{
            serversocket = factory.createServerSocket(port, clients, bindAddr);
        }
        return serversocket;
    }

    /**
     * method for intializing this factory
     * @param server the daemon of this factory
     */
    public void initialize(httpd server) {
        // Initialize instance variables:
        this.server = server ;
        this.props  = server.getProperties() ;
        this.props.registerObserver (this) ;
        // Register our property sheet:
        PropertySet set = new SocketConnectionProp("SocketConnectionProp"
                                                   , server);
        server.registerPropertySet(set);
        // Initialize parameters from properties:
        this.minFree    = props.getInteger(MINSPARE_FREE_P, MINSPARE_FREE);
        this.maxFree    = props.getInteger(MAXSPARE_FREE_P, MAXSPARE_FREE);
        this.maxIdle    = props.getInteger(MAXSPARE_IDLE_P, MAXSPARE_IDLE);
        this.maxClients = props.getInteger(MAXCLIENTS_P, MAXCLIENTS);
        String bindAddrName = props.getString(BINDADDR_P, null);
        if (bindAddrName != null) {
            try {
                bindAddr = InetAddress.getByName(bindAddrName);
            } catch (Exception ex) {
                // nothing, fallback to default
            }
        }
        // Create the LRU lists:
        idleList = new SyncLRUList();
        freeList = new SyncLRUList();
        // Create the full client list:
        csList = new IDMSocketClientState();
        if(debug)System.out.println("INF: IDMSocketClientFactory: Create all the Clients...");
        // Create all our clients:
        for (int i = 0 ; i < maxClients ; i++) {
            if ( addClient(true) == null )
                throw new RuntimeException (this.getClass().getName()
                                            + "[construstructor]"
                                            + ": unable to create clients.");
        }
        // Create the thread cache:
        if(debug)System.out.println("INF: IDMSocketClientFactory: Create the ThreadCache...");
        threadcache = new ThreadCache(server.getIdentifier() +
                                      "-socket-clients");
        threadcache.setCachesize(props.getInteger(MAXTHREADS_P, MAXTHREADS));
        threadcache.setThreadPriority(server.getClientThreadPriority());
        threadcache.setIdleTimeout(props.getInteger(IDLETO_P,IDLETO));
        threadcache.setGrowAsNeeded(true);
        threadcache.initialize();
        // Start the debugging thread, if needed:
        if ( debugthread ) {
            new DebugThread(this).start();
        }
        if(debug) System.out.println("INF: IDMSocketClientFactory: Create the SSONETServerSocketFactory.");

        // Create the SSONETContextFactory
        factory = new SSONETServerSocketFactory((SSONETContextFactory)this);
        // Create Certificate Listener
        certificateListener = new DefaultCertificateListener(this);
    }

    /**
     * Creates a SSONETContext (Method of SSONETContextFactory)
     */
    public SSONETContext createContext(){
        long times[] = new long[5];
        if(debug) times[0] = System.currentTimeMillis();
        // Create a context using role "WebServer"
        Context context = Server.idManager.createContext("Webserver");
        context.setOwnActionConfiguration(new ActionConfiguration(3,3,1,1));
        // find the pseudonym for this context
        if(debug) times[1] = System.currentTimeMillis();
        ChainEvaluator.evaluate("Pseudonymverwaltung",(ChainEvaluationListener)this,
                                null, null, context);
        // register listeners
        context.addCertificateListener(certificateListener);
        context.addConnectionListener(this);

        if(debug) System.out.println("INF: IDMSocketContextFactory: Context created.");
        if(debug){
            times[2] = System.currentTimeMillis();
            System.out.println("INF: IDMSocketClientFactory.createContext: Times: create Context: "+(times[1]-times[0])+"\n"
                              +"                                           find pseudonym to use: "+(times[2]-times[1]));
        }
        return context;
    }

    /**
     * Handles a solvable conflict during security goal negotiation
     */
    public boolean decideSecurityGoal(int securityGoal,
                                      ActionConfiguration ownConfiguration,
                                      SSONETContext context){
        // always give in
        return true;
    }

    /**
     * Receive warnings from ChainEvaluator for finding the Pseudonym for receiving
     * requests.
     */
    public boolean doWarning(String chainname, Condition cond, String id,
                             byte[] content, Context context, String parameter){
        if(debug) System.out.println("WAR: IDMSocketClientFactory: While finding the Pseudonym:\n"
            +"ChainName: "+chainname+", id: "+id+", parameter: "+parameter);
        // default: ignore the warning
        return true;
    }

    /**
     * Create a new IDMSocketClient for the pool.
     * @param free A boolean, if <strong>true</strong> the client is inserted
     * straight into the free list, otherwise, it is not plugged into any
     * list.
     * @return A SocketClientState instance, if creation of a new client was
     * allowed, <strong>null</strong> if no more clients could be created.
     */

    protected synchronized IDMSocketClientState addClient (boolean free) {
        // Create a new client.
        csList                   = new IDMSocketClientState(csList);
        IDMSocketClientState cs     = csList;
        IDMSocketClient      client = new IDMSocketClient(server, this, cs);
        cs.client = client;
        clientCount++;
        clientEstim++;
        // Plug into free LRU if required:
        if ( free ) {
            cs.status = IDMSocketClientState.C_FREE;
            freeList.toHead(cs);
            freeCount++;
        }
        return cs ;
    }


 //------------------------------------------------------------------------------
 // Methods and variables taken from SocketClientFactory
 // (We need to take all of SocketClientFactory, because the SocketClient must be changed, and
 //  the method creating SocketClients needs private variables from all over SocketClientFactory) :-(

    private static final boolean debugstats  = false;
    private static final boolean debugthread = false;

    public static final int MINSPARE_FREE = 5;
    public static final int MAXSPARE_FREE = 10;
    public static final int MAXSPARE_IDLE = 20;
    public static final int MAXTHREADS    = 40;
    public static final int MAXCLIENTS    = 32;
    public static final int IDLETO        = 10000;

    public static final int AVG_LIGHT = 1;
    public static final int AVG_NORMAL = 2;
    public static final int AVG_HIGH = 3;
    public static final int AVG_DEAD = 4;

    /** 
     * Property name for the minimum number of free clients, which must be available.
     */ 
    public final static String
    MINSPARE_FREE_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.minFree";
    /**
     * Property name for the maximum of free clients, which are allowed.
     */
    public final static String
    MAXSPARE_FREE_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.maxFree";
    /**
     * Property name for the maximum of idle (connected, but waiting for new requests)
     * clients, which are allowed.
     */
    public final static String
    MAXSPARE_IDLE_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.maxIdle";
    /**
     * Property name for the maximum number of threads, which might be used for Jigsaw.
     */
    public final static String
    MAXTHREADS_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.maxThreads";
    /**
     * Property name for the maximum number of clients allowed.
     */
    public final static String
    MAXCLIENTS_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.maxClients";
    /**
     * Property name for the the timeout for idle clients, after which 
     * the connection is closed.
     */
    public final static String
    IDLETO_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.idleTimeout";
    /**
     * Property name for internet address, where jigsaw is bound to.
     */
    public final static String
    BINDADDR_P = "org.w3c.jigsaw.http.socket.SocketClientFactory.bindAddress";

    int minFree    = 0;
    int maxFree    = 0;
    int maxIdle    = 0;

    int                  count     = 0 ;	// number of created clients.
    ObservableProperties props     = null ;
    int                  busyCount = 0 ;	// number of busy clients.

    LRUList idleList = null;
    LRUList freeList = null;

    IDMSocketClientState csList = null;

    int idleCount   = 0 ;
    int freeCount   = 0 ;
    int clientCount = 0 ;
    int clientEstim = 0 ;

    ThreadCache threadcache = null;

    int loadavg = AVG_LIGHT;

    boolean alive = true;

    /** X--X
     * Give the status of this class as a partial HTML text which will be added
     * into a block level element
     * @return a String, the generated HTML
     */
    public String getHTMLStatus() {
        int idle = 0;
        int free = 0;
        int used = 0;
        int bndc = 0;
        int reqc = 0;
        int bnd, req;
        StringBuffer sb = new StringBuffer();
        IDMSocketClientState cs = null;
        StringBuffer sb1 = null;
        if (debugstats) {
            sb1 = new StringBuffer();
        }
        // used clients
        cs = csList;
        if (debugstats) {
            sb1.append("<table border=\"1\" class=\"idle\">\n"
                       + "<caption>Used Clients"
                       +"</caption><tr><th>Id<th>BindCount<th>ReqCount<th>Diff"
                       + "<th>BoundTo</tr>\n");
        }
        while (cs != null) {
            if (cs.status == IDMSocketClientState.C_BUSY) {
                InetAddress ia = cs.client.getInetAddress();
                bnd = cs.client.getBindCount();
                req = cs.client.getRequestCount();
                if (debugstats) {
                    sb1.append ("<tr><td>" +cs.id+ "<td>"+ bnd
                                + "<td>" + req + "<td>"+ (req - bnd)+ "<td>" +
                                ((ia == null) ? "Unbound" :
                                 cs.client.getInetAddress().getHostAddress()) +
                                "</tr>\n" );
                }
                used++;
                bndc += bnd;
                reqc += req;
            }
            cs = cs.csnext;
        }
        if (debugstats) {
            sb1.append("</table>\n");
        }
        // idle clients
        cs = (IDMSocketClientState) idleList.getHead();
        if (debugstats) {
            sb1.append("<table border=\"1\" class=\"idle\">\n<caption>Idle "
                       +"Clients</caption><tr><th>Id<th>BindCount<th>ReqCount"
                       +"<th>Diff<th>BoundTo</tr>\n");
        }
        while (cs != null) {
            InetAddress ia = cs.client.getInetAddress();
            idle++;
            bnd = cs.client.getBindCount();
            req = cs.client.getRequestCount();
            if (debugstats) {
                sb1.append ("<tr><td>" +cs.id+ "<td>"+ bnd
                            + "<td>" + req + "<td>" + (req - bnd) + "<td>" +
                            ((ia == null) ? "Unbound" :
                             cs.client.getInetAddress().getHostAddress()) +
                            "</tr>\n" );
            }
            bndc += bnd;
            reqc += req;
            cs = (IDMSocketClientState)idleList.getNext(cs);
        }
        if (debugstats) {
            sb1.append("</table>\n");
        }
        // free clients
        cs = (IDMSocketClientState) freeList.getHead();
        if (debugstats) {
            sb1.append("<table border=\"1\" class=\"idle\">\n"
                       + "<caption>Free Clients"
                       + "</caption><tr><th>Id<th>BindCount<th>ReqCount<th>"
                       + "Diff</tr>\n");
        }
        while (cs != null) {
            free++;
            bnd = cs.client.getBindCount();
            req = cs.client.getRequestCount();
            if (debugstats) {
                sb1.append ("<tr><td>" + cs.id + "<td>"+ bnd
                            + "<td>" + req + "<td>" + (req - bnd) + "\n");
            }
            bndc += bnd;
            reqc += req;
            cs = (IDMSocketClientState)freeList.getNext(cs);
        }
        if (debugstats) {
            sb1.append("</table>\n");
        }

        // stats
        sb.append("<table border class=\"thread\">\n<caption>Thread counts"
                  + "</caption><tr><th>free<th>idle<th>used"
                  + "<th>estim<th>total<th>Load</tr>");
        sb.append("<tr><td>"+freeCount+"("+free+")"+"<td>"+idleCount
                  +"("+idle+")"+"<td>"+ (clientCount - freeCount - idleCount)
                  +"("+used+")"+"<td>"+clientEstim+
                  "<td>"+clientCount+"<td>"+loadavg+"</tr></table>\n");
        // usage stats
        sb.append("<table border class=\"usage\">\n<caption>Usage</caption>"
                  + "<tr><th>ReqCount<th>BindCount<th>Diff</tr>\n");
        sb.append("<tr><td>"+reqc+"<td>"+bndc+"<td>"+(reqc -bndc)+"</table>");
        if (debugstats) {
            sb.append(sb1);
        }

        if (debugstats) {
            cs = csList;
            sb.append("<table border=\"1\" class=\"idle\">\n<caption>General"
                      + " Status</caption><tr><th>Id<th>Client<th>"
                      + "Status<th>marked<th>Thread</tr>\n");
            while (cs != null) {
                sb.append ("<tr><td>" +cs.id+ "<td>" +
                           ((cs.client == null) ? "None" : "bound") +
                           "<td>" );
                switch (cs.status) {
                case IDMSocketClientState.C_IDLE:
                    sb.append ("Idle");
                    break;
                case IDMSocketClientState.C_BUSY:
                    sb.append ("Busy");
                    break;
                case IDMSocketClientState.C_FREE:
                    sb.append ("Free");
                    break;
                case IDMSocketClientState.C_KILL:
                    sb.append ("Kill");
                    break;
                case IDMSocketClientState.C_FIN:
                    sb.append ("Fin");
                    break;
                }
                sb.append ("<td>" + cs.marked);
                if (cs.client != null) {
                    sb.append("<td>" + cs.client.thread +"</tr>\n");
                } else {
                    sb.append("<td>No CLient</tr>\n");
                }
                cs = cs.csnext;
            }
            sb.append("</table>\n");
        }
        return sb.toString();
    }

    /**
     * method for handling a dynamic property modification
     * @param name the name of the property modified
     * @return true if and only if the modification has been handled successfully
     */
    public boolean propertyChanged(String name) {
        if ( name.equals(MINSPARE_FREE_P) ) {
            minFree = props.getInteger(MINSPARE_FREE_P, minFree);
        } else if ( name.equals(MAXSPARE_FREE_P) ) {
            maxFree = props.getInteger(MAXSPARE_FREE_P, maxFree);
        } else if ( name.equals(MAXSPARE_IDLE_P) ) {
            maxIdle = props.getInteger(MAXSPARE_IDLE_P, maxIdle);
        } else if ( name.equals(MAXTHREADS_P) ) {
            int maxThreads = props.getInteger(MAXTHREADS_P, -1);
            if ( maxThreads > 0 )
                threadcache.setCachesize(maxThreads);
        } else if ( name.equals(IDLETO_P) ) {
            int idleto = props.getInteger(IDLETO_P, -1);
            if ( idleto > 0 ) {
                threadcache.setIdleTimeout(idleto);
            }
        } else if ( name.equals(MAXCLIENTS_P) ) {
            int newmax = props.getInteger(MAXCLIENTS_P, -1);
            if ( newmax > maxClients ) {
                for (int i = maxClients-newmax; --i >= 0; )
                    addClient(true);
            } else if ( newmax > 0 ) {
                maxClients = newmax;
            }
        } else if (name.equals(BINDADDR_P)) {
            try {
                bindAddr = InetAddress.getByName(props.getString(BINDADDR_P,
                                                                 null));
            } catch (Exception ex) {
                // nothing
            }
        }
        return true ;
    }

    /** X--X
     * Remove this client state from the glohbal client list.
     * @param cs The client state to remove from the list.
     */

    protected synchronized void deleteClient(IDMSocketClientState cs) {
        if ( cs.csprev == null ) {
            csList = cs.csnext;
        } else if ( cs.csnext == null ) {
            cs.csprev.csnext = null;
        } else {
            cs.csprev.csnext = cs.csnext;
            cs.csnext.csprev = cs.csprev;
        }
    }


    /** X--X
     * Update our idea of the current load.
     * The one important invariant here, is that whenever the free list
     * becomes empty, then the load average should be equals to
     * <strong>AVG_DEAD</strong>. This ensures that the server will start
     * dropping connections.
     *
     */
    private final void updateLoadAverage() {
        int oldavg = loadavg;
        if ( freeCount >= maxFree ) {
            loadavg = AVG_LIGHT;
        } else if ((freeCount >= minFree) || (idleCount >= maxIdle)) {
            if ((loadavg = AVG_NORMAL) < oldavg) {
                server.thread.setPriority(Thread.MAX_PRIORITY);
            }
        } else if ( freeCount > 0 ) {
            /* idleCount < MINSPARE_IDLE */
            if ((loadavg = AVG_HIGH) > oldavg)
                server.thread.setPriority(server.getClientThreadPriority()-2);
        } else {
            loadavg = AVG_DEAD;
        }
    }

    //X--X
    private final synchronized void decrClientCount() {
        --clientCount;
        updateLoadAverage();
    }

    //X--X
    private final synchronized boolean incrFreeCount() {
        if ( clientEstim > maxClients ) {
            clientEstim--;
            return false;
        }
        ++freeCount;
        updateLoadAverage();
        return true;
    }

    //X--X
    private final synchronized boolean decrFreeCount() {
        if ( freeCount > 0 ) {
            --freeCount;
            updateLoadAverage();
            return true;
        } else {
            return false;
        }
    }

    private final synchronized boolean incrIdleCount() {
        if ((loadavg > AVG_HIGH) || (idleCount + 1 >= maxIdle))
            return false;
        ++idleCount;
        updateLoadAverage();
        return true;
    }

    private final synchronized boolean decrIdleCount() {
        if ( idleCount > 0 ) {
            --idleCount;
            updateLoadAverage();
            return true;
        } else {
            return false;
        }
    }

    /** X--X
     * Removes an idle client from the list, updates only the idle list
     * as the free count has already be accessed
     * @param client the socket to remove from the idle list
     */
    protected boolean idleClientRemove(IDMSocketClient client) {
        // If the client pool has shut down, exit straight:
        if ( ! alive )
            return false;
        IDMSocketClientState cs = client.state;
        synchronized (cs) {
            switch(cs.status) {
            case IDMSocketClientState.C_IDLE:
                decrIdleCount();
                idleList.remove(cs);
                cs.status = IDMSocketClientState.C_FREE;
                break;
            case IDMSocketClientState.C_KILL:
            case IDMSocketClientState.C_BUSY:
            default:
                break;
            }
        }
        return true;
    }

    /** X--X
     * Notify that this client has finished with its connection.
     * If the pool wants the client to be freed (because it has too many of
     * them), it makes the client kill itself (which will trigger a call to
     * the clientFinished method, were enventual cleanup is performed).
     * @param client The client that is done with its connection.
     */

    protected boolean clientConnectionFinished (IDMSocketClient client) {
        // If the client pool has shut down, exit straight:
        if ( ! alive )
            return false;
        IDMSocketClientState cs = client.state;
        synchronized (cs) {
            switch(cs.status) {
            case IDMSocketClientState.C_IDLE:
                decrIdleCount();
                idleList.remove(cs);
                break;
            case IDMSocketClientState.C_BUSY:
            case IDMSocketClientState.C_KILL:
                break;
            case IDMSocketClientState.C_FREE:
                // already freeed?
                if (client.done) {
                    client.done = false;
                    return true;
                }
            default:
                break;
            }
            if (incrFreeCount()) {
                if ( debug )
                    System.out.println(client+": now free.");
                cs.status = IDMSocketClientState.C_FREE;
                freeList.toTail(cs);
                return true;
            } else {
                if ( debug )
                    System.out.println(client+": terminate.");
                return false;
            }
        }
    }

    /** X--X
     * Notify that this client has been killed.
     * @param client The client that has terminate.
     */
    protected void clientFinished (IDMSocketClient client) {
        // If we're not alive any more, skip:
        if ( ! alive )
            return;
        IDMSocketClientState cs = client.state;
        synchronized (cs) {
            if ( debug )
                System.out.println(client+": finished "+cs.status);
            // Otherwise, perform the job:
            switch(cs.status) {
            case IDMSocketClientState.C_IDLE:
                break;
            case IDMSocketClientState.C_FREE:
                decrFreeCount();
                freeList.remove(cs);
                break;
            case IDMSocketClientState.C_BUSY:
            default:
                String msg = (client
                              + ": finished with unknown status "
                              + cs.status);
                server.errlog(msg);
                break;
            }
            cs.status = IDMSocketClientState.C_FIN;
            decrClientCount();
            deleteClient(cs);
        }
    }

    /** X--X
     * The client notifies the pool that is has been activated.
     * The client state object is updated to unmark the client as idle.
     * <p>This method needs not be synchronized, as it affect only the client
     * state, <em>not</em> the client list.
     * @param client The activated client.
     */

    protected void notifyUse(IDMSocketClient client) {
        if ( debug )
            System.out.println(client+": used.");
        IDMSocketClientState cs = client.state;
        synchronized (cs) {
            if (cs.status == IDMSocketClientState.C_IDLE) {
                decrIdleCount();
                idleList.remove(cs);
            }
            cs.status = IDMSocketClientState.C_BUSY;
        }
    }

    /** X--X
     * The client notifies the pool that it enters idle state.
     * <p>This method needs not be synchronized, as it affect only the client
     * state, <em>not</em> the client list.
     * @param client The client that is going to be idle.
     */

    protected boolean notifyIdle(IDMSocketClient client) {
        IDMSocketClientState cs = client.state;
        if ( alive ) {
            synchronized (cs) {
                if ( incrIdleCount() ) {
                    if ( debug )
                        System.out.println(client+": idle, keep-alive.");
                    cs.status = IDMSocketClientState.C_IDLE;
                    idleList.toHead(cs);
                    return true;
                } else {
                    if ( debug )
                        System.out.println(client+": idle, closed.");
                    // Kill some old idle connections, give a chance for next:
                    int killsome = Math.max((maxFree - freeCount),
                                            (maxIdle - idleCount));
                    killSomeClients((killsome > 0) ? killsome : 1);
                    // And give it a change if the load is not too high
                    if ( incrIdleCount() ) {
                        if ( debug )
                            System.out.println(client+": idle, keep-alive.");
                        cs.status = IDMSocketClientState.C_IDLE;
                        idleList.toHead(cs);
                        return true;
                    }
                    return false;
                }
            }
        } else {
            if ( debug )
                System.out.println(client+": idle (dead), closed.");
            // Kill some old idle connections, give a chance for next:
            int killsome = Math.max((maxFree - freeCount),
                                    (maxIdle - idleCount));
            killSomeClients((killsome > 0) ? killsome : 1);
            return false;
        }
    }

    //X--X
    protected void killSomeClients(int howmany) {
        int count = (howmany > 0) ? howmany : Math.max((maxFree - freeCount),
                                                       (maxIdle - idleCount));
        if (debug) {
            System.out.println("Killing :" + howmany);
        }
        while ( --count >= 0 ) {
            IDMSocketClientState cs = (IDMSocketClientState) idleList.removeTail();
            if ( cs != null ) {
                synchronized (cs) {
                    if (cs.status == IDMSocketClientState.C_IDLE) {
                        if ( debug )
                            System.out.println(cs.client +
                                               ": kill (some-client).");
                        decrIdleCount();
                        cs.status = IDMSocketClientState.C_KILL;
                        cs.client.unbind();
                    }
                }
            } else {
                break;
            }
            // if the load falls back to normal operation, we are all set
            if ((freeCount > minFree) && (idleCount < maxIdle)) {
                break;
            }
        }
    }

    final protected void killSomeClients() {
        killSomeClients(-1);
    }

    protected void run(IDMSocketClient client) {
        if ( debug )
            System.out.println(client+": warming up...");
        boolean threaded = threadcache.getThread(client, true);
        if ( debug )
            System.out.println(client+": threaded="+threaded);
    }

    /** X--X
     * Handle the given connection.
     * Find a free client, bind it to the given socket, and run it. If we
     * have reached our maximum allowed number of clients, kill some
     * connections.
     * <p>A client enters the LRU list (and become a candidate for kill) only
     * after it has handle one request (or if some timeout expires). This is
     * performed by the first call to <code>notifyUse</code> which will
     * silently insert the client into the LRU if it was not there already.
     * <p>This client pool does a lot of nice thinigs, but could probably be
     * implemented in a much better way (while keeping the features it has).
     * Contention arond the pool is probably concern number 1 of performances.
     * @param socket The connection to handle.
     */
    public void handleConnection (Socket socket) {
        if ( debug )
            System.out.println("new connection.");
        IDMSocketClientState cs = null;
        switch(loadavg) {
          case AVG_LIGHT:
              // Free list is non empty, be fast:
              if ( decrFreeCount() )
                  cs = (IDMSocketClientState) freeList.removeTail();
              break;
          case AVG_NORMAL:
          case AVG_HIGH:
              // Free list is non empty, but we try killing a client:
              killSomeClients();
              if ( decrFreeCount() ) {
                  cs = (IDMSocketClientState) freeList.removeTail();
              }
              break;
          case AVG_DEAD:
              break;
        }
        if ( debug )
            System.out.println("load "+loadavg
                               + ", client="
                               + ((cs != null)
                                  ? cs.client.toString()
                                  : "unbound"));
        // At this point, we do have a free client, bind it:
        if ( cs != null ) {
            if ( debug )
                System.out.println(cs.client+": bound.");
            cs.status = IDMSocketClientState.C_BUSY;
            cs.client.bind(socket);
        } else {
            if ( debug )
                System.out.println("*** connection refused (overloaded).");
            try {
                socket.close();
            } catch (IOException ex) {
            }
            server.errlog(socket.getInetAddress()+" refused (overloaded).");
        }
        return;
    }

    //X--X
    protected synchronized void killClients(boolean force) {
        alive = false;
        // Kill all clients (first shot):
        IDMSocketClientState cs = csList;
        while ((cs != null) && (cs.client != null)) {
            synchronized (cs) {
                // Only if a client is idely read'ing its socket, we close it
                cs.client.kill(cs.status==IDMSocketClientState.C_IDLE);
            }
            cs = cs.csnext;
        }
        // Kill all clients (second shot):
        // Some client may be in transition during first shot, second shot
        // really kills everything.
        try {
            Thread.sleep(5000);
        } catch (Exception ex) {
        }
        cs = csList;
        while ((cs != null) && (cs.client != null)) {
            synchronized (cs) {
                cs.client.kill(true);
            }
            cs = cs.csnext;
        }
    }

    /**
     * Shutdown the client pool.
     * If force is <strong>true</strong>, kill all running clients right
     * now, otherwise, wait for them to terminate gracefully, and return
     * when done.
     * @param force Should we interrupt running clients.
     */

    //X--X
    public void shutdown (boolean force) {
        // First stage: kill all clients (synchronized)
        killClients(force) ;
        // Second stage (unsynchronized), join all client threads
        IDMSocketClientState cs = csList;
        while ((cs != null) && (cs.client != null)) {
            if ( debug )
                System.out.println(cs.client+": join.");
            cs.client.join();
            cs = cs.csnext;
        }
        // Some cleanup (helps the GC, and make sure everything is free)
        props.unregisterObserver(this);
        props    = null;
        csList   = null;
        freeList = null;
        idleList = null;
        server   = null;
    }

	/* (non-Javadoc)
	 * @see org.ssonet.net.ConnectionListener#negotiationRequestReceived(org.ssonet.net.SSONETContext, int)
	 */
	public void negotiationRequestReceived(SSONETContext context, int negotiationType) {
	}

	/* (non-Javadoc)
	 * @see org.ssonet.net.ConnectionListener#connectionStateChanged(org.ssonet.net.SSONETContext, int)
	 */
	public void connectionStateChanged(SSONETContext context, int currentState) {
	}

}
