/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package cebit.server.idm.socket;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * A very specific class to buffer output sent back to the client.
 * None of the Java base class are adequat, in particular,
 * BufferedOutputStream has to inconvenient:
 * <ul>
 * <li>It enforces buffer reallocation on each new connection.
 * <li>It is not very smart with flushing the stream (it flushes it whenever
 * output exceeds buffer size, which triggers a reply header flush before
 * any data bytes can be added to the packet).
 */
class SocketOutputBuffer extends FilterOutputStream {
    protected byte buf[] = null;
    protected int  count = 0;

    public void close()
        throws IOException
    {
        try {
            try { flush();     } catch (Exception ex) {}
            try { out.close(); } catch (Exception ex) {}
        } finally {
            out   = null;
            count = 0;
        }
    }

    public void flush()
        throws IOException
    {
        if ( count > 0 ) {
            out.write(buf, 0, count);
            out.flush();
            count = 0;
        }
    }

    public void write(byte b[])
        throws IOException
    {
        write(b, 0, b.length);
    }

    public void write(byte b[], int off, int len)
        throws IOException
    {
        int avail = buf.length - count;
        if ( len < avail ) {
            System.arraycopy(b, off, buf, count, len);
            count += len;
            return;
        } else if ((avail > 0) && (count > 0)) {
            System.arraycopy(b, off, buf, count, avail);
            count += avail;
            flush();
            out.write(b, off+avail, len-avail);
        } else {
            flush();
            out.write(b, off, len);
        }
    }

    public void write(int b)
        throws IOException
    {
        if ( count == buf.length )
            flush();
        buf[count++] = (byte) b;
    }

    public void reuse(OutputStream out) {
        this.out   = out;
        this.count = 0;
    }

    public SocketOutputBuffer(OutputStream out) {
        this(out, 512);
    }

    public SocketOutputBuffer(OutputStream out, int size) {
        super(out);
        this.buf   = new byte[size];
        this.count = 0;
    }
}
