/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package org.w3c.www.protocol.http;

import idman.mngt.Context;
import idman.rules.*;

import java.io.*;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.Socket;
import java.security.cert.X509Certificate;

import org.ssonet.net.*;
import org.ssonet.net.impl.SSONETClientSocket;

import org.w3c.www.http.HttpStreamObserver;
import org.w3c.www.mime.MimeParser;
import org.w3c.www.mime.MimeParserFactory;

import cebit.proxy.Proxy;

public class HttpIDMConnection
	extends HttpBasicConnection
	implements
		HttpStreamObserver,
		SSONETContextListener,
		ChainEvaluationListener,
		CertificateListener,
		ConnectionListener {

	public static boolean debug = false;
	public static boolean debugProfiling = false;

	/** The connection context used in the id manager */
	public Context context = null;
	private boolean isContextRegistered = false;
	private String role = null;
	/** the default certificate listener */
	public CertificateListener certificateListener = null;

	/**
	 * The physical socket underlying the connection.
	 */
	private Socket socket = null;

	/**
	 * Print this connection into a String.
	 * @return A String containing the external representation for the
	 * connection.
	 */

	public String toString() {
		return inetaddr + ":" + port + "[" + id + "/" + role + "] " + th;
	}

	/**
	 * The entity stream we observe has reached its end.
	 * Notify the server that it can now reuse the connection safely for some
	 * other pending requests.
	 * @param in The stream that has reached its end of file.
	 */

	public synchronized void notifyEOF(InputStream in) {
		markIdle(false);
	}

	/**
	 * The entity stream we were to observe refuse to be observed.
	 * The connection will not be reusable, so we should detach it from the
	 * managing server, without closing it, since the entity reader will
	 * close it itself.
	 * @param in The stream that has been closed.
	 */

	public synchronized void notifyFailure(InputStream in) {
		markIdle(true);
	}

	/**
	 * The entity stream we observe has been closed.
	 * After making sure the entire entity has been read, we can safely hand
	 * out the connection to the server, for later reuse.
	 * @param in The stream that has been closed.
	 */

	public synchronized void notifyClose(InputStream in) {
		try {
			byte buffer[] = new byte[1024];
			while (in.read(buffer) > 0);
			markIdle(false);
		} catch (IOException ex) {
			markIdle(true);
		}
	}

	/**
	 * Mark this connection as being used.
	 * The server, which keeps track of idle connections, has decided to use
	 * this connection to run some request. Mark this connection as used
	 * and unregister it from the server's list of idle connections.
	 * <p>Some assumptions are checked before handing out the connection
	 * for use, which can throw an RuntimeException.
	 * @return A boolean, <strong>true</strong> if the connection can be used
	 * or reused, <strong>false</strong> otherwise (the connection was detected
	 * idle, and destroy itself).
	 * @exception RuntimeException If the connection is in an invalid state.
	 */

	public synchronized boolean markUsed() {
		cached = false;
		if (debug)
			System.out.println(this +" used !");
		if (th != null)
			throw new RuntimeException(this +" already used by " + th);
		th = Thread.currentThread();
		// TODO: check if socket is alive, set to null if not
		//if ( ! socket.isAlive())/////////////////
		//   socket = null;///////////////////////
		if (socket != null) {
			cached = true;
			// do new negotiation of security
			try {
				if (context != null) {
					if (debug)
						System.out.println(
							"HttpIDMConnection.markUsed: Perform new Negotiation.");
					if (false)
						context.doSecurityGoalNegotiation();
				} else {
					if (debug)
						System.out.println(
							"HttpIDMConnection.markUsed: Implemenation Error. Context has to be != null.");
					System.exit(-1);
				}
			} catch (IOException ioe) {
				if (debug) {
					System.out.println(
						"HttpIDMConnection.markUsed: error in subsequent negotiation.");
					ioe.printStackTrace();
				}
				// Close that connection (cleanup):
				close();
				// Mark that connection as dead:
				 ((HttpBasicServer) server).unregisterConnection(this);
				((HttpBasicServer) server).deleteConnection(this);
				return false;
			}
		}
		// create a new socket
		if (socket == null) {
			if (debug)
				System.out.println(
					"HttpIDMConnection.markUsed: Socket is null.");
			try {
				try {
					long times[] = new long[4];
					if (debugProfiling)
						times[0] = System.currentTimeMillis();
					// create a ssonet socket
					context = createContext(role);
					if (debugProfiling)
						times[1] = System.currentTimeMillis();
					if (debug)
						System.out.println(
							"HttpIDMConnection.markUsed: try create SSONETSocket.");
					socket =
						new SSONETClientSocket(
							inetaddr,
							port,
							(SSONETContext) context);
					context.setTargetAddress(inetaddr);
					context.setTargetPort(port);
					if (debugProfiling)
						times[2] = System.currentTimeMillis();
					if (debug)
						System.out.println(
							"HttpIDMConnection.markUsed: SSONETSocket created.");
					socket.setSoTimeout(timeout);
					output = new BufferedOutputStream(socket.getOutputStream());
					input = new BufferedInputStream(socket.getInputStream());
					if (debugProfiling) {
						times[3] = System.currentTimeMillis();
						System.out.println(
							"HttpIDMConnection: Times for createContext: "
								+ (times[1] - times[0])
								+ "\n"
								+ "                       create SSONETSocket: "
								+ (times[2] - times[1])
								+ "\n"
								+ "                  get in- and outputStream: "
								+ (times[3] - times[2]));
					}
				} catch (ConnectException ce) {
					throw ce;
					// no system exit in this case! might happen very often
				} catch (Exception e) {
					// implementation error exception
					e.printStackTrace();
					System.exit(-1);
				}

				// initialize th input parser
				parser = new MimeParser(input, reply_factory);
				if (debug)
					System.out.println(
						"HttpIDMConnection.markUsed: MimeParser created.");
			} catch (Throwable ex) {
				// Close that connection (cleanup):
				close();
				// Mark that connection as dead:
				 ((HttpBasicServer) server).unregisterConnection(this);
				((HttpBasicServer) server).deleteConnection(this);
				return false;
			}
		}
		if (debug)
			System.out.println(
				"HttpIDMConnection.markUsed: try server.unregisterConnection.");
		synchronized (server) {
			notifyAll();
		}
		((HttpBasicServer) server).unregisterConnection(this);
		synchronized (server) {
			notifyAll();
		}
		if (debug)
			System.out.println(
				"EEE HttpIDMConnection.markUsed: server.unregisterConnection finished.");
		return true;
	}

	/**
	 * The connection is now idle again.
	 * Mark the connection as idle, and register it to the server's list of
	 * idle connection (if this connection can be reused). If the connection
	 * cannot be reused, detach it from the server and forget about it (the
	 * caller will close it by closing the entity stream).
	 * @param close Should this connection be physically closed (it is not
	 * reusable), or should we try to keep track of it for later reuse.
	 * @exception RuntimeException If the connection is in an invalid state.
	 */

	public synchronized void markIdle(boolean close) {
		// Has this connection already been marked idle ?
		if (th == null)
			return;
		if (debug)
			System.out.println(this +" idle !" + close);
		// Check consistency:
		// if ( Thread.currentThread() != th )
		//    throw new RuntimeException(this +
		//	 		         " th mismatch " +
		//		 	         th +
		//			         "/" +
		//			         Thread.currentThread());
		// Ok, mark idle for good:
		th = null;
		// close the socket, if neccessary
		if (close) {
			if (debug)
				System.out.println(this +" closing !");
			close();
			((HttpBasicServer) server).deleteConnection(this);
		} else {
			// Notify the server that a new connection is available:
			if (debug)
				System.out.println(
					"HttpIDMConnection.markIdle: try HttpBasicServer.registerConnection.");
			((HttpBasicServer) server).registerConnection(this);
			if (debug)
				System.out.println(
					"HttpIDMConnection.markIdle: HttpBasicServer.registerConnection finished.");
		}
	}

	/**
	 * Some data available on input, while writing to the server.
	 * This callback gets called when the client is emitting data to the
	 * server and the server has sent us something before we actually sent
	 * all our bytes.
	 * <p>Take any appropriate action.
	 */

	public void notifyInputAvailable(InputStream in) {
		return;
	}

	/**
	 * Get the MIME parser to read from this connection.
	 * All access to the connection's input stream should go through the MIME
	 * parser to ensure buffering coherency.
	 * @return A MimeParser instance suitable to parse the reply input stream.
	 * @exception RuntimeException If the connection was not connected.
	 */

	public MimeParser getParser() {
		if (parser == null)
			throw new RuntimeException("getParser while disconnected.");
		return parser;
	}

	/**
	 * Get the connection output stream.
	 * @return The output stream to send data on this connection.
	 * @exception RuntimeException If the connection was not previously opened.
	 */

	public OutputStream getOutputStream() {
		if (output == null)
			throw new RuntimeException("getOutputStream while disconnected.");
		return output;
	}

	/**
	 * Create a new connection.
	 * To be used only by HttpIDMServer instances.
	 */

	HttpIDMConnection(
		HttpServer server,
		int id,
		InetAddress addr,
		int port,
		int timeout,
		MimeParserFactory reply_factory)
		throws IOException {
		this(server, id, addr, port, timeout, reply_factory, "Anonym");
		new Exception().printStackTrace();
	}
	HttpIDMConnection(
		HttpServer server,
		int id,
		InetAddress addr,
		int port,
		int timeout,
		MimeParserFactory reply_factory,
		String role)
		throws IOException {
		super(server, id, addr, port, timeout, reply_factory);
		this.server = server;
		this.inetaddr = addr;
		this.port = port;
		this.id = id;
		this.timeout = timeout;
		this.reply_factory = reply_factory;
		this.role = role;
	}
	//----------------------------------- methods added for IDM
	/**
	 * Creates a context for this connection
	 */
	protected Context createContext(String role) {
		try {
			Context context = Proxy.idManager.createContext(role);
			context.setOwnActionConfiguration(
				new ActionConfiguration(3, 3, 0, 0));
			context.addCertificateListener(this);
			context.addSSONETContextListener(this);
			context.addConnectionListener(this);
			if (debug)
				System.out.println(
					"HttpIDMConnection: Context for >" + role + "< created.");
			return context;
		} catch (Exception e) {
			// Implementation Error Exceptions
			e.printStackTrace();
			System.exit(-1);
			return null;
		}
	}

	/**
	 * Check Servers certificate, register the certifiate with the id manager,
	 * and find out clients pseudonym.
	 */
	public int checkCertificate(SSONETContext context, X509Certificate cert) {
		long times[] = new long[4];
		if (debugProfiling)
			times[0] = System.currentTimeMillis();
		// create DefaultCertificateListener on first call
		if (certificateListener == null)
			certificateListener =
				new DefaultCertificateListener((ChainEvaluationListener) this);
		if (debugProfiling)
			times[1] = System.currentTimeMillis();
		// Call DefaultCertificateListener
		int ret = certificateListener.checkCertificate(context, cert);
		if (ret == CertificateListener.VALID) {
			if (debugProfiling)
				times[2] = System.currentTimeMillis();
			// find the pseudonym for this context
			ChainEvaluator.evaluate(
				"Pseudonymverwaltung",
				(ChainEvaluationListener) this,
				null,
				null,
				(Context) context);
		}
		if (debugProfiling) {
			times[3] = System.currentTimeMillis();
			System.out.println(
				"HttpIDMConnection: Time needed for create DefaultCertificateL.: "
					+ (times[1] - times[0])
					+ "\n"
					+ "                                              checkCertificate: "
					+ (times[2] - times[1])
					+ "\n"
					+ "                                              chain evaluation:  "
					+ (times[3] - times[2]));
		}
		return ret;
	}

	/**
	 * Receive warnings from ChainEvaluator for finding the Pseudonym for receiving
	 * requests.
	 */
	public boolean doWarning(
		String chainname,
		Condition cond,
		String id,
		byte[] content,
		Context context,
		String parameter) {
		if (debug)
			System.out.println(
				"IDMSocketClientFactory: Waring while finding the Pseudonym:\n"
					+ "ChainName: "
					+ chainname
					+ ", id: "
					+ id
					+ ", parameter: "
					+ parameter);
		// default: ignore the warning
		return true;
	}

	/**
	 * Handles a new negotiation request. Does nothing.
	 * @param context the <code> SSONETContext </code> used by the server for this
	 *                request.
	 * @param negotiationType <code>SECURITY_GOAL_NEGOTIATION</code> if the negotiation
	 *      requested is a security goal (and mechanisms) negotiation,<p>
	 *      <code>MECHANISMS_NEGOTIATION</code> if the negotiation requested is only a
	 *      mechanisms negotiation.
	 */
	public void negotiationRequestReceived(
		SSONETContext context,
		int negotiationType) {
		if (debug)
			System.out.println(
				"HttpIDMConnection: Negotiation request received.");
		return;
	}

	/**
	 * Handles a solvable conflict during security goal negotiation
	 */
	public boolean decideSecurityGoal(
		int securityGoal,
		ActionConfiguration ownConfiguration,
		SSONETContext context) {
		// always give in
		return true;
	}

	/**
	 * Method of SSONETContextListener. Replaces an old context.
	 */
	public void replaceContext(SSONETContext o, SSONETContext n) {
		context = (Context) n;
	}

	/**
	 * Notifies about context changes. Not used.
	 */
	public void contextChanged(SSONETContext c) {
		return;
	}

	protected Object monitor = new Object();
	public void connectionStateChanged(SSONETContext c, int state) {
		//long times[] = new long[2];
		//if(debugProfiling) times[0] = System.currentTimeMillis();
		if (state == ConnectionListener.RUNNING_STATE) {
			// register the connection
			if ((context != null) && !isContextRegistered) {
				synchronized (monitor) {
					Proxy.idManager.registerContext((Context) c);
					isContextRegistered = true;
				}
			}
		} else if (state == ConnectionListener.FINISHED_STATE) {
			// unregister context
			if ((context != null) && isContextRegistered) {
				synchronized (monitor) {
					Proxy.idManager.unregisterContext((Context) c);
					isContextRegistered = false;
				}
			}
		}
		//if(debugProfiling){
		//    times[1] = System.currentTimeMillis();
		//    System.out.println("HttpIDMConnection: Time needed for registering connection state change: "+(times[1]-times[0]));
		//}
	}

}
