/**
 * Copyright (c) 2002 Sebastian Clau, IDM Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package org.w3c.www.protocol.http;

import java.io.IOException;
import java.net.InetAddress;

import org.w3c.www.mime.MimeParserException;
import org.w3c.jigsaw.auth.AuthFilter;

/**
 * The server class, to run requests.
 * A server instance (ie an object conforming to the <code>HttpServer</code>
 * interface is resposnsible for running all requests to a given host.
 * <p>To do so, it manages a connnnection pool (ie a set of available
 * connections) which it negotiates with the global HTTP manager. It keeps
 * track of the connections it creates in order to serialize requests
 * to the target server (when possible). In order to avoid deadlock due
 * to implicit ordering of requests, a server that manages persistent
 * (ie > 1.0) connections will always be able to open at least two
 * connections to the same target.
 * <p>Connections are kept track of by the server instance, which maintains
 * at all point in time a list of <em>idle</em> connections. The way this is
 * done may seem quite tricky, but beleive me, it's the best way I have
 * found to handle it (there is here a typical deadlock situation - the same
 * which is found in org.w3c.jigsaw.http.Client
 * - due to the fact that the server instances need to be locked in order
 * to change the connection pool, and at the same time each connection of
 * the pool might have to notify the pool  of some events. This can easily
 * produce a deadlock...This situation is avoided here, by having the server
 * locked only when appropriate...
 * <p>The first HttpIDMServer initializes the identity management object. This is
 * static to this class, so all objects have access to it, and all connections
 * can be registered with the identity manager.
 */
public class HttpIDMServer extends HttpBasicServer {
    public static boolean debug = false;

    /**
     * Empty constructor.
     */
    HttpIDMServer() {
        super();
    }

    /**
     * Connections management - Allocate a new connection for this server.
     * The connection is bound to the next available IP address, so that
     * we are able to round-robin on them. If one of the DNS advertized
     * IP address fails, we just try the next one, until one of them
     * succeed or all of them fail.
     * <p>If this method has to create a new connection, a HttpIDMConnection is
     * created.
     * @return A freshly allocated connection, inserted in the idle connection
     * list.
     * @exception IOException If the server is unreachable through all of
     * Its IP addresses.
     */
    protected synchronized HttpBasicConnection allocateConnection()
        throws IOException
    {
        new Exception().printStackTrace();
        return allocateConnection("Anonym");
    }
    protected synchronized HttpBasicConnection getConnection()
        throws IOException
    {
        new Exception().printStackTrace();
        return getConnection("Anonym");
    }
    protected HttpBasicConnection getConnection(String role)
        throws IOException
    {
        while ( true ) {
            // Ask for a spare connection first:
            HttpBasicConnection conn = null;
            while ( true ) {
                if(debug) System.out.println("HttpIDMServer.getConnection: try getConnection.");
                conn = (HttpBasicConnection) manager.getConnection(this);
                if(debug) System.out.println("HttpIDMServer.getConnection: got connection.");
                if ( conn == null ){
                    if(debug) System.out.println("HttpIDMServer.getConnection: conn is null.");
                    break;
                }
                try {
                    if(debug) System.out.println("HttpIDMServer.getConnection: check if input is available.");
                    conn.input.available();
                    if(debug) System.out.println("HttpIDMServer.getConnection: input is available.");
                } catch (IOException ex) {
                    // mark that connection as dead
                    unregisterConnection(conn);
                    deleteConnection(conn);
                    conn.close();
                    if(debug) System.out.println("HttpIDMServer.getConnection: connection defect. closed.");
                    break;
                }
                synchronized (this) {
                    if(debug) System.out.println("HttpIDMServer.getConnection: reuse old connection:"+conn);
                    if ( conn.markUsed() )
                        return conn;
                }
            }
            // Negotiate the creation of a new connection:
            if(debug) System.out.println("HttpIDMServer.getConnection: try manager.negotiateConnection.");
            if ( manager.negotiateConnection(this) ) {
                if(debug) System.out.println(Thread.currentThread().toString()+" HttpIDMServer.getConnection: try allocateConnection.");
                conn = allocateConnection(role);
                if(debug) System.out.println(Thread.currentThread().toString()+" HttpIDMServer.getConnection: allocateConnection finished.");
                synchronized (this) {
                    if ( conn.markUsed() ) {
                        return conn;
                    } else {
                        // Failed to establish a fresh connection !
                        return null;
                    }
                }

            }
            // Wait for a connection to become available:
            try {
                if (debug) System.out.println("HttpIDMServer.getConnection: wait for a connection.");
                if (debug) System.out.println("\t"+state.conns);
                manager.waitForConnection(this);
            } catch (InterruptedException ex) {
                // interrupted, probably a timeout -> fail
                return null;
            }
        }
    }

    protected synchronized HttpBasicConnection allocateConnection(String role)
        throws IOException
    {
        // Create the new connection, try until all ip addrs have been checked
        if(debug)System.out.println("HttpIDMServer.allocateConnection: started.");
        HttpBasicConnection conn = null;
        try {
            if(debug)System.out.println("HttpIDMServer.allocateConnection: try updateHostAddr.");
            updateHostAddr();
            if(debug)System.out.println(">>> HttpIDMServer.allocateConnection: updateHostAddr finished.");
        } catch (HttpException hex) {
            // unable to update, will continue with the old version
            // as it may be a temporary DNS problem
            if(debug)System.out.println("HttpIDMServer.allocateConnection: no addrs available.");
            if (addrs == null) {
                throw new IOException(hex.getMessage());
            }
        }
        if (debug) System.out.println("HttpIDMServer.allocateConnection: looping adresses.");
        for (int loop = 0 ; loop < addrs.length ; loop++) {
            InetAddress addr = null;
            int         i    = addrptr;
            do {
                if ((addr = addrs[i]) != null)
                    break;
                i = (i+1) % addrs.length;
            } while ( i != addrptr );
            addrptr = (addrptr+1) % addrs.length;
            if (addr == null) {
                throw new IOException("Host "+host+" resolved to a null"
                                      + " InetAddr");
            }
            try {
                if(debug) System.out.println("HTTPIDMServer.allocateConnection: Try create new HttpIDMConnection.");
                conn = new HttpIDMConnection(this
                                               , connid++
                                               , addr
                                               , port
                                               , timeout
                                               , manager.getReplyFactory()
                                               , role);
                if(debug) System.out.println(">>> HTTPIDMServer.allocateConnection: new HttpIDMConnection created.");                                               
                break;
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        // If successfull, register the connection as one of ours:
        if ( conn == null ){
            if(debug) System.out.println("EEE HTTPIDMServer.allocateConnection: conn is null :-( .");
            throw new IOException("Unable to connect to "+host);
        }
        if(debug) System.out.println("HttpIDMServer.allocateConnection: try manager.notifyConnection.");
        manager.notifyConnection(conn);
        if(debug) System.out.println("EEE HttpIDMServer.allocateConnection: return new connection.");
        return conn;
    }

    /**
     * Run the given request in synchronous mode.
     * <p> This method is taken from <code>HttpbasicConnection</code>. It differs
     * only for the fact, that it always marks connections as idle, if they are
     * not explicitely detected as keep-alive connections.
     * @param request The request to run.
     * @return A Reply instance, containing the reply headers, and the
     * optional reply entity, to be read by the calling thread.
     * @exception HttpException If the request processing failed.
     */

    public Reply runRequest(Request request)
        throws HttpException
    {
        RequestObserver o = request.getObserver();
        if ( debug )
            System.out.println(">>> Running request: "+request.getURL());
        // Acquire a connection to run the request:
        HttpBasicConnection conn  = null;
        Reply               reply = null;
        try {
            String role = (String)request.getState(AuthFilter.STATE_AUTHUSER);
            if (debug) System.out.println("HttpIDMServer.runRequest: request: "+request);
            if (debug) System.out.println("HttpIDMServer.runRequest: role: "+role);
            // Notify that request is being queued:
            if(debug) System.out.println("HttpIDMServer.runRequest: notifyObserver(1).");
            if ( o != null )
                notifyObserver(o, request, RequestEvent.EVT_QUEUED);
            // Allocate a connection and run the request:
            int maxretry = 3;
            for (int i = 0 ; (reply == null) && (i < maxretry) ; i++) {
                // getConnection can return null, check documentation:
                if(debug) System.out.println("HttpIDMServer.runRequest: try getConnection.");
                if ((conn  = getConnection(role)) == null)
                    continue;
                // Some connection ready, try using it:
                if(debug) System.out.println("HttpIDMServer.runRequest: try request.hasOutputStream.");
                if (request.hasOutputStream()) {
                    if(debug) System.out.println("HttpIDMServer.runRequest: try request.getContentLength.");
                    int length = request.getContentLength();
                    if(debug) System.out.println("HttpIDMServer.runRequest: request.outputStream.mark.");
                    if (length != -1) {
                        request.getOutputStream().mark(length+1);
                    } else {
                        request.getOutputStream().mark(1024*64);
                    }
                }
                try {
                    if(debug) System.out.println("HttpIDMServer.runRequest: try doRequest:"+request.getURL()
                        +", with conn:"+conn.toString());
                    if ((reply = doRequest(conn, request)) == null) {
                        // if a cached connection fails, it may be due
                        // to a socket in bad shape...
                        // workaround for the missing socket.isAlive() method
                        if (conn.cached)
                            --i;
                        // restore the body if ther was one.
                        if (request.hasOutputStream()) {
                            request.getOutputStream().reset();
                        }
                        // Get rid of that useless connection:
                        conn.markIdle(true);
                    }
                } catch (MimeParserException mex) {
                    if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+": MimeParserException.");
                    throw (mex);
                } catch (Throwable ioex) {
                    if (debug)
                        ioex.printStackTrace();
                    // connection 1.0 failed so it was not a cached connection
                    // restore the body if ther was one.
                    try {
                        if (request.hasOutputStream()) {
                            request.getOutputStream().reset();
                        }
                    } catch (IOException mrex) {
                        // mark/reset not supported on the stream
                    }
                    // Get rid of that useless connection:
                    conn.markIdle(true);
                }
            }
            // Did the request really failed ?
            if ( reply == null ) {
                System.out.println("HttpIDMServer.runRequest:"+conn.toString()+": Reply is null.");
                String msg = ("Unable to contact target server "
                              + this
                              + " after " + maxretry + " tries.");
                if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+":  notifyObserver(2).");
                if ( o != null )
                    notifyObserver(o, request, RequestEvent.EVT_UNREACHABLE);
                throw new HttpException(request, msg);
            }
            if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+":  reply.matchesRequest.");
            reply.matchesRequest(request);
            if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+":  notifyObserver(3).");
            if ( o != null )
                notifyObserver(o, request, RequestEvent.EVT_REPLIED);
            // Can we try to reuse the connection ?
            if ( reply.keepsAlive() ) {
                if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+":  reply.setStreamObserver.");
                reply.setStreamObserver(conn);
                // If the reply has no input stream, register the connection:
                // else it may be HTTP-CONNECT oder Switching protocols.. don't reuse this, its still in use
                if ( ! reply.hasInputStream() )
                    conn.markIdle(false);
            } else {
                if(debug) System.out.println("HttpIDMServer.runRequest:"+conn.toString()+":  deleteConnection.(keep = false)");
                deleteConnection(conn);
            }
            if ( debug ){
                System.out.println(""+conn.toString()+": Request done !");
            }
        } catch (IOException ex) {
            ex.printStackTrace();
            if ( conn != null )
                conn.markIdle(true);
            if ( o != null )
                notifyObserver(o, request, RequestEvent.EVT_CLOSED);
            throw new HttpException(request, ex);
        } catch (MimeParserException ex) {
            ex.printStackTrace();
            if ( conn != null )
                conn.markIdle(true);
            if ( o != null )
                notifyObserver(o, request, RequestEvent.EVT_CLOSED);
            throw new HttpException(request, ex);
        }
        return reply;
    }


}

